package com.htfp.weather.download;

import com.htfp.weather.info.Constant;
import com.htfp.weather.utils.DateTimeUtils;
import com.htfp.weather.utils.HttpClientUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.context.annotation.DependsOn;
import org.springframework.stereotype.Component;
import ucar.nc2.NetcdfFile;
import ucar.nc2.NetcdfFiles;

import javax.annotation.Resource;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.time.*;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;


/**
 * @author shiyi
 */
@Slf4j
@Component("gfsDownloader")
@DependsOn("gfsDataConfig")
public class GfsDownloader extends BaseDataDownloader {
    @Resource
    GfsDataConfig gfsDataConfig;
    private final ExecutorService executorService = Executors.newFixedThreadPool(5);
    private int hourDiff;
    private OffsetDateTime realUpdateDateTime;

    public GfsDownloader(GfsDataConfig gfsDataConfig) {
        super();
        this.gfsDataConfig = gfsDataConfig;
    }

    /**
     * 初始化下载起始时间
     */
    public void iniTimeSetting() {
        // 当前时间
        this.setStartTime(LocalDateTime.now());
        this.setStartTimeStr(LocalDateTime.now().format(DateTimeFormatter.ofPattern(Constant.BJT_TIME_STRING)));
        // 以下时间均为UTC时间
        OffsetDateTime utcStartTime = DateTimeUtils.getUTCDateTime(OffsetDateTime.now());
        realUpdateDateTime = getAvailableUpdateDateTime(utcStartTime);
        this.setRefTime(realUpdateDateTime);
        this.setRefTimeStr(realUpdateDateTime.format(DateTimeFormatter.ofPattern(Constant.UTC_TIME_STRING)));
        // 起报点和实际请求时间的差异

        this.hourDiff = (int) Duration.between(realUpdateDateTime, utcStartTime).toHours();
        log.info("[GFS Download] 起始为当前本地时间：{}", this.getStartTime());
        log.info("[GFS Download] GFS最近更新时间UTC：{}，本地时间: {}", realUpdateDateTime, DateTimeUtils.getLocalZoneDateTime(realUpdateDateTime));
    }

    @Override
    public boolean download(FileInfo fileInfo) {
        try {
            String url = fileInfo.getUrl();
            File destDir = new File(fileInfo.getSavePath());
            if (!destDir.exists()) {
                destDir.mkdirs();
            }
            File fileOut = new File(destDir, fileInfo.getFileName());
            // if (fileOut.exists()) {
            //     log.info("[GFS Download] 文件已存在，忽略下载: {}", fileOut.getAbsolutePath());
            //     return true;
            // }
            log.info("[GFS Download] 文件下载中，保存至 {}", fileOut);
            // FileOutputStream fos = new FileOutputStream(filePath);
            // try {
            //     ReadableByteChannel rbc = Channels.newChannel(new URL(url).openStream());
            //     fos.getChannel().transferFrom(rbc, 0, Long.MAX_VALUE);
            // } catch (ConnectException e) {
            //     log.error("连接超时:{}", url);
            // }
            //
            // fos.flush();
            // fos.close();
            FileUtils.copyURLToFile(new URL(url), fileOut);
            log.info("[GFS Download] 文件下载成功: {}", fileOut);
            return fileValid(fileOut.getAbsolutePath());
        } catch (Exception e) {
            log.error("[GFS Download] 文件 {} 下载失败：{}", fileInfo, e.getMessage());
            return false;
        }
    }
    private boolean fileValid(String file) {
        try (NetcdfFile ncFile = NetcdfFiles.open(file)) {
            ncFile.getLocation();
            return true;
        } catch (IOException e){
            return false;
        }
    }
    @Override
    public boolean downloadAll(List<FileInfo> fileInfoList) {
        log.info("[GFS Download] 下载任务启动，共 {} 个文件", fileInfoList.size());
        List<Future<Boolean>> futures = new ArrayList<>();
        for (FileInfo fileInfo : fileInfoList) {
            futures.add(executorService.submit(() -> download(fileInfo)));
        }
        boolean allCompleted = true;
        for (Future<Boolean> future : futures) {
            try {
                allCompleted = allCompleted && future.get();
            } catch (InterruptedException | ExecutionException e) {
                e.printStackTrace();
            }
        }
        return allCompleted;
    }

    @Override
    public List<FileInfo> getFilesInfo() {
        // 时间选项填充
        String lonlatBoxStr = null;
        String levelsStr = null;
        String variablesStr = null;
        try {
            lonlatBoxStr = fillUrlForLonLat();
            levelsStr = fillUrlForLevels();
            variablesStr = fillUrlForVariables();
        } catch (Exception e) {
            log.error("[GFS Download] 文件下载参数生成错误");
            throw new RuntimeException(e);
        }

        // 预报数据时间分辨率
        final int forecastStep = 1;
        // 分辨率为3h，整数截断；包括目标时刻本身
        final int nFiles = 1 + gfsDataConfig.getDuration() / forecastStep;
        // 存储文件信息
        List<FileInfo> fileInfoList = new ArrayList<>(nFiles);
        for (int i = 0; i < nFiles; i++) {
            FileInfo fileInfo = new FileInfo();
            int forecastHour = i * forecastStep + hourDiff;
            fileInfo.setForecastHour(forecastHour);
            String baseURL = getBaseURL(forecastHour, levelsStr, variablesStr, lonlatBoxStr);
            fileInfo.setUrl(baseURL);
            fileInfo.setRefTimeStr(this.getRefTime().format(DateTimeFormatter.ofPattern(Constant.UTC_TIME_STRING)));
            fileInfo.setForecastUTCTimeStr(
                    this.getRefTime().plusHours(fileInfo.getForecastHour())
                            .format(DateTimeFormatter.ofPattern(Constant.UTC_TIME_STRING))
            );
            fileInfo.setForecastBJTimeStr(
                    DateTimeUtils.getLocalZoneDateTime(this.getRefTime().plusHours(fileInfo.getForecastHour()))
                            .format(DateTimeFormatter.ofPattern(Constant.BJT_TIME_STRING))
            );
            fileInfo.setFileName(String.format("%s-from-%s+%d.grib2", fileInfo.getForecastBJTimeStr(), fileInfo.getRefTimeStr(), forecastHour));
            fileInfo.setSavePath(gfsDataConfig.getSaveRoot() + "/" + fileInfo.getRefTimeStr());
            fileInfoList.add(fileInfo);
        }
        return fileInfoList;
    }


    private String getBaseURL(int forcastHour, String levelsStr, String variablesStr, String lonlatBoxStr) {
        String baseFormat = "https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25_1hr.pl?dir=/gfs.%s/%02d/atmos&file=gfs.t%02dz.pgrb2.0p25.f%03d";

        String realUpdateDateString = realUpdateDateTime.format(DateTimeFormatter.ofPattern("yyyyMMdd"));
        int refHour = realUpdateDateTime.getHour();

        String baseurl = String.format(baseFormat, realUpdateDateString, refHour, refHour, forcastHour);
        StringJoiner stringJoiner = new StringJoiner("&");

        return stringJoiner.add(baseurl).add(levelsStr).add(variablesStr).add(lonlatBoxStr).toString();
    }

    /**
     * 获取距离指定时刻最近的更新时刻
     */
    private OffsetDateTime getAvailableUpdateDateTime(OffsetDateTime utcStartTime) {


        // 获取临近的更新点
        int hour = utcStartTime.getHour();
        int[] updateHours = {0, 6, 12, 18};
        int nearUpdateHour = 0;
        for (int i : updateHours) {
            if (i < hour) {
                nearUpdateHour = i;
            }
        }
        OffsetDateTime nearUpdateDateTime = utcStartTime.withHour(nearUpdateHour).withMinute(0).withSecond(0).withNano(0);

        OffsetDateTime now = OffsetDateTime.now();
        // 如果最近的更新点还没有更新数据，那么返回前一个更新点
        // if (isGfsUpdated(nearUpdateDateTime)) {
        if (now.isAfter(nearUpdateDateTime.plusHours(4))) {
            // GFS延迟4小时才能获取到数据，比如04时更新00时的数据
            return nearUpdateDateTime;
        } else {
            return nearUpdateDateTime.minusHours(6);
        }

    }

    /**
     * 判断GFS是否更新
     *
     * @return
     */
    public boolean isGfsUpdated(OffsetDateTime utcUpdateTime) {
        String timeString = utcUpdateTime.format(DateTimeFormatter.ofPattern("'gfs.'yyyyMMdd/HH"));
        String url = String.format("https://nomads.ncep.noaa.gov/pub/data/nccf/com/gfs/prod/%s/", timeString);
        try {
            String httpString = HttpClientUtils.sendGet(url, null);
            if (StringUtils.isEmpty(httpString)) {
                return false;
            }
            // FIXME 2024/5/17: 可能遇到正在更新，文件夹存在，但是文件不一定存在
            return httpString.contains("atmos/") && httpString.contains("wave/");
        } catch (Exception e) {
            return false;
        }
    }

    private String fillUrlForLonLat() {
        // 根据传入参数生成对应的URL请求信息，生成字符串
        Double topLat = gfsDataConfig.getMaxLat();
        Double leftLon = gfsDataConfig.getMinLon();
        Double rightLon = gfsDataConfig.getMaxLon();
        Double bottomLat = gfsDataConfig.getMinLat();
        return String.format("subregion=&toplat=%.2f&leftlon=%.2f&rightlon=%.2f&bottomlat=%.2f", topLat, leftLon, rightLon, bottomLat);
    }

    private String fillUrlForLevels() {
        int[] pressureLevels =  gfsDataConfig.getPressureLevels();
        if (pressureLevels.length == 0) {
            throw (new IllegalArgumentException("至少指定一个层次"));
        }
        StringJoiner stringJoiner = new StringJoiner("&");
        for (int pres : pressureLevels) {
            stringJoiner.add(String.format("lev_%d_mb=on", pres));
        }
        // 添加地面2m和10m的信息
        int[] heightLevels = gfsDataConfig.getHeightLevels();
        for (int height : heightLevels) {
            stringJoiner.add(String.format("lev_%d_m_above_ground", height));
        }
        return stringJoiner.toString();
    }

    private String fillUrlForVariables() {
        ArrayList<String> variables = gfsDataConfig.getVariables();
        StringJoiner stringJoiner = new StringJoiner("&");
        if (variables.isEmpty()) {
            throw (new IllegalArgumentException("至少指定一个下载变量"));
        }
        for (String variable : variables) {
            stringJoiner.add(String.format("var_%s=on", variable));
        }
        return stringJoiner.toString();
    }

}
