package com.htfp.weather.web.service;

import com.aliyun.tablestore.grid.consts.AttributionEnum;
import com.aliyun.tablestore.grid.model.GridDataSet;
import com.aliyun.tablestore.grid.model.GridDataSetMeta;
import com.aliyun.tablestore.grid.model.grid.Grid4D;
import com.htfp.weather.griddata.common.GfsVariableNameEnum;
import com.htfp.weather.griddata.common.TableConfigBean;
import com.htfp.weather.griddata.operation.GfsDataFetcher;
import com.htfp.weather.griddata.operation.QueryMeta;
import com.htfp.weather.info.Constant;
import com.htfp.weather.utils.DateTimeUtils;
import com.htfp.weather.utils.NdArrayUtils;
import com.htfp.weather.web.exception.AppExcpetion;
import com.htfp.weather.web.exception.ErrorCode;
import com.htfp.weather.web.pojo.response.PlaneResponse;
import com.htfp.weather.web.pojo.response.ProfileResponse;
import com.htfp.weather.web.pojo.response.TimeSeriesDataset;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;
import ucar.ma2.Array;

import javax.annotation.Resource;
import java.time.Duration;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @Author : shiyi
 * @Date : 2024/2/2 15:47
 * @Description : 获取tablestore格点数据
 */
@Service("tablestore-gfs") @Slf4j
public class GfsDataServiceImpl implements IDataService {
    @Resource
    GfsDataFetcher gfsDataFetcher;
    @Resource
    QueryMeta queryMeta;
    @Resource
    TableConfigBean tableConfigBean;


    public ProfileResponse getProfile(OffsetDateTime targetTime, String variableName, double latitude, double longitude) {
        String targetVariable = GfsVariableNameEnum.getGfsVariableName(variableName);

        if (targetVariable == null) {
            throw new AppExcpetion(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }

        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": e.getMessage()");
        }
        int iTime = getTimeIndex(targetTime, lastGridDataSetMeta);
        int iLat = getLatitudeIndex(latitude);
        int iLon = getLongitudeIndex(longitude);
        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            int[] pressureLevels = tableConfigBean.getLevList();
            Array array = gfsDataFetcher.getProfile(gridDataSetId, targetVariable, iTime, iLat, iLon);
            float[] values = (float[]) array.getStorage();
            return new ProfileResponse(pressureLevels, values);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

    }

    public TimeSeriesDataset getForecastSeries(double latitude, double longitude, int pressureLevel) {
        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": e.getMessage()");
        }
        long milli = (long) lastGridDataSetMeta.getAttributes().get(AttributionEnum.REFERENCE_TIME.getName());
        OffsetDateTime refTime = OffsetDateTime.ofInstant(Instant.ofEpochMilli(milli), ZoneOffset.ofHours(0));
        OffsetDateTime currentTime = DateTimeUtils.getUTCDateTime(OffsetDateTime.now());
        if (currentTime.isBefore(refTime)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }
        int startHour = getForecastHourFromTargetTime(currentTime, refTime);
        List<String> forecastHours = lastGridDataSetMeta.getForecastHours();
        List<Integer> iTimeList = new ArrayList<>();
        List<String> timeList = new ArrayList<>();

        for (int i = 0; i < forecastHours.size(); i++) {
            int hour = Integer.parseInt(forecastHours.get(i));
            if (hour >= startHour) {
                iTimeList.add(i);
                String timeStr = DateTimeUtils.getLocalZoneDateTime(refTime.plusHours(hour))
                        .format(DateTimeFormatter.ofPattern(Constant.API_TIME_STRING));
                timeList.add(timeStr);
            }
            if (iTimeList.size() >= 24) {
                break;
            }
        }
        int[] iTimes = iTimeList.stream().mapToInt(Integer::valueOf).toArray();
        int iLat = getLatitudeIndex(latitude);
        int iLon = getLongitudeIndex(longitude);
        int iLev = getLevelIndex(pressureLevel);
        List<String> variableNames = GfsVariableNameEnum.getVariableNamesInFile();
        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            GridDataSet gridDataSet = gfsDataFetcher.getSeries(gridDataSetId, variableNames, iTimes, iLev, iLat, iLon);
            return buildTimeSeriesDataset(timeList, gridDataSet);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public PlaneResponse getPlane(OffsetDateTime targetTime, String variableName, int level, double minLat, double maxLat, double minLon, double maxLon) {
        String targetVariable = GfsVariableNameEnum.getGfsVariableName(variableName);

        if (targetVariable == null) {
            throw new AppExcpetion(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }
        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": e.getMessage()");
        }

        int iTime = getTimeIndex(targetTime, lastGridDataSetMeta);
        int iLev = getLevelIndex(level);
        int iMinLat = getLatitudeIndex(minLat);
        int iMaxLat = getLatitudeIndex(maxLat);
        int iMinLon = getLongitudeIndex(minLon);
        int iMaxLon = getLongitudeIndex(maxLon);
        int latLength = iMaxLat - iMinLat + 1;
        int lonLength = iMaxLon - iMinLon + 1;

        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            Array array = gfsDataFetcher.getPlane(gridDataSetId, targetVariable, iTime, iLev, iMinLat, iMinLon, latLength, lonLength);

            return buildPlane(array, iMinLat, iMinLon, latLength, lonLength);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

    }

    private PlaneResponse buildPlane(Array array, int iMinLat, int iMinLon, int latLength, int lonLength) {
        PlaneResponse planeResponse = new PlaneResponse();
        float[][] values = (float[][]) array.reduce().copyToNDJavaArray();
        planeResponse.setLatList(Arrays.copyOfRange(tableConfigBean.getLatList(), iMinLat, iMinLat + latLength));
        planeResponse.setLonList(Arrays.copyOfRange(tableConfigBean.getLonList(), iMinLon, iMinLon + lonLength));
        planeResponse.setValues(values);
        return planeResponse;
    }

    private TimeSeriesDataset buildTimeSeriesDataset(List<String> timeList, GridDataSet gridDataSet) {
        TimeSeriesDataset timeSeriesDataset = new TimeSeriesDataset(timeList.size());
        timeSeriesDataset.setTime(timeList);
        Grid4D grid4D;
        grid4D = gridDataSet.getVariable(GfsVariableNameEnum.getGfsVariableName(GfsVariableNameEnum.TEMP));
        if (grid4D != null) {
            timeSeriesDataset.setTemp((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableNameEnum.getGfsVariableName(GfsVariableNameEnum.WIND_SPEED));
        if (grid4D != null) {
            timeSeriesDataset.setWindSpeed((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableNameEnum.getGfsVariableName(GfsVariableNameEnum.WIND360));
        if (grid4D != null) {
            timeSeriesDataset.setWind360((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableNameEnum.getGfsVariableName(GfsVariableNameEnum.CLOUD));
        if (grid4D != null) {
            timeSeriesDataset.setCloud((float[]) grid4D.toArray().getStorage());
        }
        return timeSeriesDataset;
    }

    private int getTimeIndex(OffsetDateTime targetTime, GridDataSetMeta lastGridDataSetMeta) {
        long milli = (long) lastGridDataSetMeta.getAttributes().get(AttributionEnum.REFERENCE_TIME.getName());
        OffsetDateTime refTime = OffsetDateTime.ofInstant(Instant.ofEpochMilli(milli), ZoneOffset.ofHours(0));
        if (targetTime.isBefore(refTime)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }

        List<String> forecastHours = lastGridDataSetMeta.getForecastHours();
        String targetHour = String.valueOf(getForecastHourFromTargetTime(targetTime, refTime));
        if (!forecastHours.contains(targetHour)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }
        int iTime = forecastHours.indexOf(targetHour);
        return iTime;
    }

    private int getLongitudeIndex(double lon) {
        double[] lonList = tableConfigBean.getLonList();
        if (lonList.length == 0) {
            throw new AppExcpetion(ErrorCode.LONGITUDE_INDEX_ERROR, "经度列表为空， 无法查找索引");
        }
        double minLon = lonList[0];
        double maxLon = lonList[lonList.length - 1];
        if (minLon > lon || lon > maxLon) {
            throw new AppExcpetion(ErrorCode.LONGITUDE_INDEX_ERROR, "经度不在有效范围内: " + minLon + " ~ " + maxLon);
        }
        int nearestIndex = NdArrayUtils.findNearestIndex(lonList, lon);
        if (nearestIndex == -1) {
            throw new AppExcpetion(ErrorCode.LONGITUDE_INDEX_ERROR, "找不到" + lon + "的经度索引");
        }
        return nearestIndex;
    }

    private int getLatitudeIndex(double lat) {
        double[] latList = tableConfigBean.getLatList();
        if (latList.length == 0) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "纬度列表为空， 无法查找索引");
        }
        double minLat = latList[0];
        double maxLat = latList[latList.length - 1];

        if (minLat > lat || lat > maxLat) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "纬度不在有效范围内: " + minLat + " ~ " + maxLat);
        }
        int nearestIndex = NdArrayUtils.findNearestIndex(latList, lat);
        if (nearestIndex == -1) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "找不到" + lat + "的纬度索引");
        }
        return nearestIndex;
    }

    private int getLevelIndex(int pressureLevel) {
        List<Integer> levList = Arrays.stream(tableConfigBean.getLevList()).boxed().collect(Collectors.toList());
        if (CollectionUtils.isEmpty(levList)) {
            throw new AppExcpetion(ErrorCode.LEVEL_INDEX_ERROR, "经度列表为空， 无法查找索引");
        }
        return levList.indexOf(pressureLevel);
    }

    private int getForecastHourFromTargetTime(OffsetDateTime targetTime, OffsetDateTime refTime) {
        return (int) Duration.between(refTime, targetTime).toHours();
    }


}
